// Solutions to the Advent Of Code 2024.
// Copyright (C) 2024  Stefan Müller
//
// This program is free software: you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#pragma once

#include <vector>

#include <aoc/Grid.hpp>
#include <aoc/GridPathData.hpp>
#include <aoc/LinesSolver.hpp>
#include <aoc/Point2.hpp>

class GuardGallivant
    : public LinesSolver
{
public:
    virtual const std::string getPuzzleName() const override;
    virtual const std::string getInputFileName() const override;
    virtual void processDataLine(const std::string& line) override;
    virtual void finish() override;
private:
    Point2 start_{};
    const size_t getStartDirectionIndex() const;
    const char getStartChar() const;
    const char getObstructionChar() const;
    void tracePath();
    void visitPosition(const Point2& current, const size_t directionIndex, Grid<GridPathData>& pathGrid);
    void backtracePath(const Point2& current, const size_t directionIndex, Grid<GridPathData>& pathGrid);
    void backtraceTurn(const Point2& current, const size_t reverseTurnDirectionIndex, Grid<GridPathData>& pathGrid);
    size_t turnDirectionRight(const size_t currentDirectionIndex) const;
    size_t turnDirectionLeft(const size_t currentDirectionIndex) const;
    size_t revertDirection(const size_t currentDirectionIndex) const;
};
